"""
Notifly Documentation

Discord: (object): Discord wrapper to send message, images to the channel using Webhooks
Slack (object): Slack wrapper to send message, images to the channel using API
Telegram (object): Telegram Web Client wrapper to send images, files over the bot using API
"""



"""Slack-API call wrapper"""

import slack
from slack.errors import SlackApiError
from notifly.tf_notifier import TfNotifier


class Notifier(TfNotifier):
    def __init__(self, token, channel='general'):
        """
        Initialize the slack webclient instance using API.

        Args:
            token (string): API token [Mandatory].
            channel (string): Post's the notification to it's corresponding channel [default = #general].
        Returns:
            The output of slack API_test which verify the Authentication protocol.
        Raises:
            SlackApiError.
        """
        self.__client = slack.WebClient(token=token)
        self.__channel = channel

        try:
            self.__client.api_test()['ok']
        except SlackApiError as api_err:
            print(f"Got an error: {api_err.response['error']}")
            exit(1)

    def send_message(self, msg) -> object:
        """
        Function to post message to the slack channel.

        Args:
            msg (string): Enter your message to post.
        Returns:
            Outputs the response of message on post operation.
        Raises:
            SlackApiError
        """
        try:
            return self.__client.chat_postMessage(channel=self.__channel, text=msg)
        except SlackApiError as api_err:
            print(f"Got an error: {api_err.response['error']}")
            exit(1)

    def send_file(self, file_path) -> object:
        """
        Function to post an file to the slack channel.

        Args:
            file_path (string): Enter the path of the file to be sent.
        Returns:
            Outputs the response of message on post operation.
        Raises:
                FileNotFoundError
        """
        try:
            return self.__client.files_upload(file=file_path, channels=self.__channel)
        except FileNotFoundError as fl_err:
            print(fl_err)



from subprocess import Popen, PIPE
from distutils import spawn
import os
import sys
import platform


def gpu():
    """
    Monitors the hardware info on the runtime.

    Returns:
        The list of parameters containing hardware info
    """
    if platform.system() == "Windows":
        try:
            nvidia_smi = spawn.find_executable('nvidia-smi')
        except FileNotFoundError as fn:
            print(fn)

        if nvidia_smi is None:
            nvidia_smi = "%s\\Program Files\\NVIDIA Corporation\\NVSMI\\nvidia-smi.exe" % os.environ['systemdrive']
    else:
        nvidia_smi = "nvidia-smi"

    # Get ID, processing and memory utilization for all GPUs (deviceIds, uuid, gpuUtil, memTotal, memUsed, memFree,
    # driver,gpu_name,serial,display_mode,display_active, temp_gpu)
    try:
        p = Popen([nvidia_smi,
                   "--query-gpu=index,uuid,utilization.gpu,memory.total,memory.used,memory.free,driver_version,name,"
                   "gpu_serial,display_active,display_mode,temperature.gpu",
                   "--format=csv,noheader,nounits"], stdout=PIPE)
        stdout, stderr = p.communicate()
        if p is not None:
            x = stdout.decode('UTF-8').split(',')
            if len(x) > 1:
                return x

        print('No Gpu Found, continuing with CPU instance')

    except SystemError as stderr:
        print('Unable to establish a communication with GPU', stderr)
        sys.exit(1)
    except FileNotFoundError as fn:
        print('Unable to find GPU connection in your device, Proceeding using CPU instance')



"""Discord API webhooks wrapper"""

import requests
from notifly.tf_notifier import TfNotifier
from requests import exceptions


class Notifier(TfNotifier):

    class __AuthError(Exception):
        """
        Authentication Exception
        """
        pass

    def __init__(self, webhooks):
        """
        Initialize the Discord Webhook instance

        Args:
            webhooks (basestring): Discord webhook token [Mandatory]
        Returns:
            The response of webhook status
        Raises:
            Exception, AuthError, MissingSchema
        """
        self.__webhooks = webhooks
        self.payload = None

        try:
            if requests.get(self.__webhooks).status_code == 401:
                raise Notifier.__AuthError('Invalid Webhook')
        except exceptions.ConnectionError as err:
            print('HTTPS Connection Error - Unable to reach the destination')
            exit(1)
        except Notifier.__AuthError as ty_err:
            print(ty_err)
            exit(1)
        except requests.models.MissingSchema as ms_err:
            print(ms_err)
            exit(1)

    def send_message(self, msg) -> object:
        """
        Function to post message to the discord channel

        Args:
            msg (string): Posts message to the discord channel [String] [UTF-8]
        Returns:
            The response to the message on post operation.
        Raises:
            ConnectionError
        """
        payload = {'content': str(msg)}
        try:
            return requests.post(url = self.__webhooks, data = payload)
        except exceptions.ConnectionError as cer:
            print(cer)
            exit(1)

    def send_file(self, file_path) -> object:
        """
        Function to post an image to the discord channel

        Args:
            file_path (string): Enter the path of the file to be sent
        Returns:
            The response to the message on post operation.
        Raises:
            FileNotFoundError, OverflowError
        """
        try:
            self.payload = {'file': open(file_path, 'rb')}
        except FileNotFoundError as fl_er:
            print(fl_er)
            exit(1)
        try:
            return requests.post(url = self.__webhooks, files = self.payload)
        except OverflowError as err:
            print('Size Overflow Error', err)
            exit(1)



""" Telegram Wrapper"""

import json
import os
import sys
import requests
from notifly.tf_notifier import TfNotifier


class Notifier(TfNotifier):
    def __init__(self, token, dir_download= './downloads'):
        """
        Initialize the telegram client using tokens to access the HTTP API

        Args:
            token (string): API token [Mandatory].
        Args:
            dir_download (string): -> Storage location for dumping payload [default = "./downloads"]
        """
        self.token = token
        self.api_url = "https://api.telegram.org/bot{}/".format(token)
        self.file_url = "https://api.telegram.org/file/bot{}/".format(token)
        self.dirDownloads = dir_download

    def __get_updates(self, offset=0, timeout=10):
        """
        Get the latest updates as json file

        Args:
            offset (int): default value 0
            timeout (int): Timeout for the request post (default value 10)
        Returns:
            The json results
        Raises:
            TimeoutError
        """
        method = 'getUpdates'
        params = {'timeout': timeout, 'offset': offset}
        resp = requests.get(self.api_url + method, params)
        try:
            return resp.json()['result']
        except KeyError:
            print('TimeoutError')
            sys.exit(1)

    def __chat_id_response(self) -> int:
        """
        Fetches the latest chat id from the client

        Returns:
            The latest chat-id of the client
        Raises:
            TimeoutError
        """
        try:
            fetch_updates = self.__get_updates()
            return fetch_updates[0]['message']['chat']['id']
        except TimeoutError as tm_err:
            print(tm_err)
            sys.exit(1)

    def send_message(self, msg, notification=False) -> object:
        """
        Function to send message

        Args:
            msg (string): Enter the message to post
            notification (bool): Disable_Notification (default=False)
        Returns:
            The status_code of the post operation (send_message)
        Raises:
            IndexError
        """
        try:
            method = 'sendMessage'
            params = {'chat_id': self.__chat_id_response(), 'text': msg,
                      'parse_mode': 'HTML', 'disable_notification': notification}
            return requests.post(self.api_url + method, params)
        except IndexError:
            print('Time out error')
            sys.exit(1)

    def send_image(self, img_path) -> object:
        """
        Function to send image via telegram

        Args:
            img_path (basestring): Enter the file_path to send the image file
        Returns:
            The status_code of the post operation (send_image)
        Raises:
            FileNotFoundError, InvalidFormatError
        """
        method = 'sendPhoto?' + 'chat_id=' + str(self.__chat_id_response())
        if img_path[-4:] not in ['.jpg', '.png']:
            print('Invalid File Format, please use .jpg or .png format')
            sys.exit(1)
        try:
            files = {'photo': open(img_path, 'rb')}
            return requests.post(self.api_url + method, files = files)
        except FileNotFoundError as fl_err:
            print(fl_err)
            sys.exit(1)

    def send_file(self, file_path) -> object:
        """
        Function to send documents via telegram

        Args:
            file_path (basestring): Enter the file_path to send the image file
        Returns:
            The status_code of the post operation (send_document)
        Raises:
            FileNotFoundError, TimeoutError
        """
        method = 'sendDocument?' + 'chat_id=' + str(self.__chat_id_response())
        try:
            files = {'document': open(file_path, 'rb')}
            return requests.post(self.api_url + method, files = files)
        except FileNotFoundError as fn_err:
            print(fn_err)
            sys.exit(1)
        except TimeoutError as tm_err:
            print(tm_err)
            sys.exit(1)

    def session_dump(self) -> json:
        """
        Function to Dump all the data from the telegram client during the current session

        Returns:
            Dumps the session details in the form of json file.
        Raises:
            IOError
        """
        resp = self.__get_updates()
        try:
            if not os.path.exists(self.dirDownloads):
                os.mkdir(self.dirDownloads)
            local_path = os.path.join(self.dirDownloads, 'session_dump.json')

            with open(local_path, 'w+', encoding='utf-8') as outfile:
                json.dump(resp, outfile)
        except IOError as io_err:
            print(io_err)
            sys.exit(1)



import ssl
from notifly import tf_notifier
import tensorflow as tf

ssl._create_default_https_context = ssl._create_unverified_context
token = 'https://discord.com/api/webhooks/771253998313144321/5FNBsAi8-exyc-3rx' \
        'LOYudQdQqMdpcaCOqCh6z1IX4ledG1oGyaONeIY1HDUns6qBZhW'
notifier = tf_notifier.TfNotifier(token=token, platform='discord')


class TestCallback(tf.keras.callbacks.Callback):

    @notifier.notify_on_epoch_begin(epoch_interval=1, graph_interval=10)
    def on_epoch_begin(self, epoch, logs=None):
        pass

    @notifier.notify_on_epoch_end(epoch_interval=1, graph_interval=10)
    def on_epoch_end(self, epoch, logs=None):
        pass

    @notifier.notify_on_train_begin()
    def on_train_begin(self, logs=None):
        pass

    @notifier.notify_on_train_end()
    def on_train_end(self, logs=None):
        pass


fashion_mnist = tf.keras.datasets.fashion_mnist

(train_images, train_labels), (test_images, test_labels) = fashion_mnist.load_data()

class_names = ['T-shirt/top', 'Trouser', 'Pullover', 'Dress', 'Coat',
               'Sandal', 'Shirt', 'Sneaker', 'Bag', 'Ankle boot']

model = tf.keras.Sequential([
    tf.keras.layers.Flatten(input_shape=(28, 28)),
    tf.keras.layers.Dense(5, activation='relu'),
    tf.keras.layers.Dense(10)
])

model.compile(optimizer='adam',
              loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
              metrics=['accuracy'])

model.fit(train_images, train_labels, epochs=5, callbacks=[TestCallback()])

test_loss, test_acc = model.evaluate(test_images, test_labels, verbose=2)

print('\nTest accuracy:', test_acc)



from notifly import gpu_stats
import inspect
import matplotlib.pyplot as plt
import copy
import psutil


class TfNotifier:

    def send_message(self, message):
        return

    def send_file(self, file_path):
        return

    @staticmethod
    def get_hardware_stats():
        """
        Fetches the hardware stats on regular intervals

        Returns:
            The parameters of CPU, GPU, RAM
        Raises:
            Exception
        """
        try:
            cpu_usage = psutil.cpu_percent(interval = 0.2)
            mem_stats = psutil.virtual_memory()
            ram_usage = round((mem_stats.used / mem_stats.total) * 100, 2)
        except Exception as e:
            print('Unable to find system parameters', e)

        x = gpu_stats.gpu()
        if x is None:
            return f"CPU Usage: {cpu_usage}%, RAM Usage: {ram_usage}%"

        gpu_util = x[2]
        tot_v_ram = x[3]
        v_ram_used = x[4]
        unused_vram = x[5]
        driver_ver = x[6]
        gpu_name = x[7]
        gpu_temp = x[11].strip()
        return f"CPU Usage: {cpu_usage}%, RAM Usage: {ram_usage}%, GPU Usage: {gpu_util}%, GPU Temp: {gpu_temp}," \
               f" GPU Memory: {v_ram_used} MB, GPU Unused Memory: {unused_vram} MB"

    @staticmethod
    def plot_graph(history, current_epoch_logs):
        """
        Plots the graphs on runtime

        Args:
            history (dict): gets the logs of the tensorflow model.train()
            current_epoch_logs (int): Fetches the epochs on runtime event
        Returns:
            Plots the graph on runtime
        """

        # merge history with current epoch logs
        for key, value in current_epoch_logs.items():
            if key not in history:
                history[key] = [value]
            else:
                history[key].append(value)

        plt.figure()
        for key, value in history.items():
            if len(value) != 1:
                plt.plot(range(1, len(value) + 1), value, label = str(key))
            else:
                plt.scatter(range(1, len(value) + 1), value, label = str(key))
            plt.title("Training History")

        plt.xlabel('Epochs')
        plt.legend()
        plt.pause(1e-13)
        file_path = 'fig.png'
        plt.savefig(file_path, bbox_inches = 'tight')
        plt.close()
        return file_path

    def notify_on_train_begin(self):
        """
        Decorator which runs on beginning of the training model.

        Returns:
            Updated function
        """

        def inner(func_to_call):
            def wrapper(*args, **kwargs):
                # get parameter values
                parameter_values = list(args)
                for i in kwargs.values():
                    parameter_values.append(i)

                # get arguments names from the function signature
                sig = inspect.signature(func_to_call)
                parameter_names = sig.parameters.keys()

                # parameter mapping with names to values
                parameter_mapping = dict(zip(parameter_names, parameter_values))

                # get starting logs
                starting_logs = parameter_mapping.get('logs')

                self.send_message(f':muscle: training started, got log keys: {starting_logs}')

                return func_to_call(*args, **kwargs)

            return wrapper

        return inner

    def notify_on_train_end(self):
        """
        Decorator which runs on end of each iteration during training model.

        Returns:
            Updated function
        """
        def inner(func_to_call):
            def wrapper(*args, **kwargs):
                # get parameter values
                parameter_values = list(args)
                for i in kwargs.values():
                    parameter_values.append(i)

                # get arguments names from the function signature
                sig = inspect.signature(func_to_call)
                parameter_names = sig.parameters.keys()

                # parameter mapping with names to values
                parameter_mapping = dict(zip(parameter_names, parameter_values))

                # get starting logs
                ending_logs = parameter_mapping.get('logs')

                self.send_message(f':tada: training ended, got log keys: {ending_logs}')

                return func_to_call(*args, **kwargs)

            return wrapper

        return inner

    def notify_on_epoch_begin(self, epoch_interval, graph_interval, hardware_stats_interval):
        """
        Decorator which runs on beginning of the training model.

        Returns:
            Updated function
        """
        def inner(func_to_call):
            def wrapper(*args, **kwargs):

                # get parameter values
                parameter_values = list(args)
                for i in kwargs.values():
                    parameter_values.append(i)

                # get arguments names from the function signature
                sig = inspect.signature(func_to_call)
                parameter_names = sig.parameters.keys()

                # parameter mapping with names to values
                parameter_mapping = dict(zip(parameter_names, parameter_values))

                # get model instance
                model_instance = parameter_mapping.get('self').model

                # get current epoch because epoch starts from 0
                current_epoch = parameter_mapping.get('epoch') + 1

                # get current epoch logs
                current_epoch_logs = parameter_mapping.get('logs')

                # notify if current_epoch is divisible by epoch_interval
                if current_epoch % epoch_interval == 0:
                    message = f"epoch: {current_epoch} started, got log keys:"
                    for k, v in current_epoch_logs.items():
                        message += " {}: {:.4f} ".format(k, v)
                    self.send_message(message)

                # notify if current_epoch is divisible by hardware_stats_interval
                if current_epoch % hardware_stats_interval == 0:
                    hardware_stats = TfNotifier.get_hardware_stats()
                    self.send_message(hardware_stats)

                # notify graph if current_epoch is divisible by graph_interval
                if current_epoch % graph_interval == 0:
                    history_copy = copy.deepcopy(model_instance.history.history)
                    file_path = TfNotifier.plot_graph(history_copy, current_epoch_logs)
                    # TODO: change this function call
                    self.send_file(file_path)

                return func_to_call(*args, **kwargs)

            return wrapper

        return inner

    def notify_on_epoch_end(self, epoch_interval, graph_interval, hardware_stats_interval):
        """
        Decorator which runs on end of the training model.

        Returns:
            Updated function
        """
        def inner(func_to_call):
            def wrapper(*args, **kwargs):

                # get parameter values
                parameter_values = list(args)
                for i in kwargs.values():
                    parameter_values.append(i)

                # get arguments names from the function signature
                sig = inspect.signature(func_to_call)
                parameter_names = sig.parameters.keys()

                # parameter mapping with names to values
                parameter_mapping = dict(zip(parameter_names, parameter_values))

                # get model instance
                model_instance = parameter_mapping.get('self').model

                # get current epoch because epoch starts from 0
                current_epoch = parameter_mapping.get('epoch') + 1

                # get current epoch logs
                current_epoch_logs = parameter_mapping.get('logs')

                # notify if current_epoch is divisible by epoch_interval
                if current_epoch % epoch_interval == 0:
                    message = f"epoch: {current_epoch} ended, got log keys:"
                    for k, v in current_epoch_logs.items():
                        message += " {}: {:.4f} ".format(k, v)
                    self.send_message(message)

                # notify if current_epoch is divisible by hardware_stats_interval
                if current_epoch % hardware_stats_interval == 0:
                    hardware_stats = TfNotifier.get_hardware_stats()
                    self.send_message(hardware_stats)

                # notify graph if current_epoch is divisible by graph_interval
                if current_epoch % graph_interval == 0:
                    history_copy = copy.deepcopy(model_instance.history.history)
                    file_path = TfNotifier.plot_graph(history_copy, current_epoch_logs)
                    # TODO: change this function call
                    self.send_file(file_path)

                return func_to_call(*args, **kwargs)

            return wrapper

        return inner



import ssl
from notifly import discord
import tensorflow as tf
import os
from dotenv import load_dotenv

load_dotenv()

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'


def test():
    ssl._create_default_https_context = ssl._create_unverified_context
    token = os.getenv('TOKEN')
    notifier = discord.Notifier(token)

    class TestCallback(tf.keras.callbacks.Callback):

        @notifier.notify_on_epoch_begin(epoch_interval=1, graph_interval=1, hardware_stats_interval=1)
        def on_epoch_begin(self, epoch, logs=None):
            pass

        @notifier.notify_on_epoch_end(epoch_interval=1, graph_interval=1, hardware_stats_interval=1)
        def on_epoch_end(self, epoch, logs=None):
            pass

        @notifier.notify_on_train_begin()
        def on_train_begin(self, logs=None):
            pass

        @notifier.notify_on_train_end()
        def on_train_end(self, logs=None):
            pass

    fashion_mnist = tf.keras.datasets.fashion_mnist
    (train_images, train_labels), (_, _) = fashion_mnist.load_data()

    model = tf.keras.Sequential([
        tf.keras.layers.Flatten(input_shape=(28, 28)),
        tf.keras.layers.Dense(2, activation='relu'),
        tf.keras.layers.Dense(10)
    ])

    model.compile(optimizer='adam',
                  loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
                  metrics=['accuracy'])

    model.fit(train_images, train_labels, epochs=1, callbacks=[TestCallback()])


test()
